import qubovert
import pyxis as px
import numpy as np
from sklearn.metrics.pairwise import cosine_distances

from qubo_nn.data import LMDBDataLoader


QUBO_SIZE = 9


def solve_qubo2(item):
    Q = qubovert.utils.matrix_to_qubo(item.reshape(QUBO_SIZE, QUBO_SIZE))
    sol = Q.solve_bruteforce(all_solutions=False)
    sol_ = [0 for _ in range(QUBO_SIZE)]
    for k, v in sol.items():
        sol_[k] = v
    return sol_


cfg = {
    'dataset_id': 'a3dbg_all',
    'use_big': False,
    'model': {
        'batch_size': 1,
        'train_eval_split': 1.0,
        'shuffle_data': False
    }
}

lmdb_loader = LMDBDataLoader(cfg)
loader = lmdb_loader.train_data_loader
data = list(loader)

labels = [d[1][0].numpy() for d in data]
data = [d[0][0].numpy() for d in data]
data_copy = [d.copy() for d in data]

for x in data:
    print(np.percentile(np.diagonal(x, 0, 1, 2), 50, axis=0))
# import pdb; pdb.set_trace()

sols = [np.array([solve_qubo2(d) for d in curr_data]) for curr_data in data]

def make_similar(our_data, their_data, sols, factor=.1):
    cached_data = our_data.copy()

    diff = their_data - our_data
    our_data += diff * factor

    new_sol = np.array([solve_qubo2(d) for d in our_data])
    print(abs(new_sol - sols).sum())
    binary_err = 0
    real_err = 0
    total_val = 0
    for i, (new_sol_, sol) in enumerate(zip(new_sol, sols)):
        best_val = sol.T @ cached_data[i] @ sol
        new_val = new_sol_.T @ cached_data[i] @ new_sol_

        binary_err += min(
            abs(new_sol_ - sol).sum(),
            abs((1 - new_sol_) - sol).sum()
        )
        real_err += abs(best_val - new_val)
        total_val += abs(best_val)
    return our_data, binary_err, real_err, total_val


factors = [0] + [.6] * 10

prob = ["NP", "MC", "MVC", "SP", "M2SAT", "SPP", "QA", "QK", "M3SAT", "TSP", "MCQ"]

total_len = QUBO_SIZE * len(data[0])
data_og = data_copy[0]  # The OG. Basically the problem we make all others similar to.
for i in range(11):
    if i == 0:
        new_data, binary_err, real_err, total_val = make_similar(data_og, data_copy[1], sols[i], factors[i])
    else:
        new_data, binary_err, real_err, total_val = make_similar(data_copy[i], data_og, sols[i], factors[i])
    data[i] = new_data
    print(prob[i], binary_err, binary_err / total_len, real_err, real_err / total_val)
    

data_flattened = [d.flatten() for d in data]
print("COS DIST")
print(cosine_distances(data_flattened).round(2))


def save(data, labels, cfg_id):
    data_new = []
    for d in data:
        data_new.extend(d)
    data = np.array(data_new)
    labels_new = []
    for label in labels:
        labels_new.extend(label)
    labels = np.array(labels_new)
    dirpath = 'datasets/%s/'
    db = px.Writer(
        dirpath=dirpath % cfg_id,
        map_size_limit=60000,
        ram_gb_limit=60
    )
    db.put_samples('input', data, 'target', labels)
    db.close()


cfg_id = "sim_mult_%d_%d" % tuple(int(100 * f)  for f in factors[:2])
save(data, labels, cfg_id)
print(cfg_id)



import copy
import json


class Config:
    def __init__(self, base_path=''):
        with open(base_path + 'simulations.json', 'r') as f:
            self.cfg = json.load(f)

    def _update_cfg(self, base_cfg, new_cfg):
        # We support one level for now.
        for k in new_cfg.keys():
            if k == 'base_cfg' or k == 'desc' or k == 'dataset_id':
                continue
            base_cfg[k].update(new_cfg[k])
        base_cfg['dataset_id'] = new_cfg['dataset_id']

    def get_cfg(self, cfg_id):
        if cfg_id not in self.cfg:
            raise Exception(
                'Error: Key %s does not exist in simulations.json.' % cfg_id
            )

        initial_base_cfg = self.cfg["1"]
        base_cfg = self.cfg[self.cfg[cfg_id].get('base_cfg', cfg_id)]
        # All base configs are based on config "1".
        # This enables backwards compatibility when new options are added.
        self._update_cfg(initial_base_cfg, base_cfg)

        cfg = copy.deepcopy(initial_base_cfg)
        self._update_cfg(cfg, self.cfg[cfg_id])

        cfg['cfg_id'] = cfg_id
        return cfg


if __name__ == '__main__':
    cfg = Config()
    print('\n'.join(cfg.cfg.keys()))



import random

import networkx as nx
import matplotlib.pyplot as plt
from dwave_qbsolv import QBSolv
from networkx.generators.random_graphs import gnm_random_graph

from qubo_nn.problems.max_cut import MaxCut


def solve_qubo(qubo):
    qb = QBSolv()

    Q = {}
    # We can assume a quadratic matrix.
    for i in range(qubo.shape[0]):
        for j in range(qubo.shape[1]):
            if qubo[i][j] != 0:
                Q[(i, j)] = qubo[i][j]
    response = qb.sample_qubo(Q, num_repeats=1000)
    print(response)
    ret = [0] * len(response.samples()[0])
    for k, v in response.samples()[0].items():
        ret[k] = v
    return ret


def modify_qubo1(qubo):
    for i in range(len(qubo)):
        for j in range(len(qubo)):
            if qubo[i][j] == 0:
                if random.random() > 0.9:
                    qubo[i][j] = 1.
    return qubo


G = gnm_random_graph(16, 20, seed=123)
# nx.draw(G)
# plt.show()

mc = MaxCut({}, G)
qubo = -mc.gen_qubo_matrix() + 0
print(qubo)
qubo = modify_qubo1(qubo)
print(qubo)

solution = solve_qubo(qubo)
print(solution)
assert\
    solution == [0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0] or\
    solution == [1, 1, 1, 1, 1, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1]



import gzip
import json
import socket
import random
import pickle
import datetime
import itertools

import torch
import qubovert
import pyxis as px
import numpy as np
import networkx as nx
import qubo_nn.problems
from qubo_nn.nn import Optimizer
from qubo_nn.nn import ReverseOptimizer
from qubo_nn.nn import AutoEncoderOptimizer
from qubo_nn.nn import RNNOptimizer
from qubo_nn.nn import A3Optimizer
from qubo_nn.nn import Resistance1
from qubo_nn.nn import Resistance2
from qubo_nn.nn import QbsolvOptimizer
from qubo_nn.nn import RedAEOptimizer
from qubo_nn.logger import Logger
from qubo_nn.problems import PROBLEM_REGISTRY
from qubo_nn.data import LMDBDataLoader


np.set_printoptions(suppress=True)


class Classification:
    def __init__(self, cfg):
        self.cfg = cfg
        self.chunks = cfg['problems']['chunks']
        self.n_problems = cfg['problems']['n_problems']
        self.qubo_size = cfg['problems']['qubo_size']
        self.scramble_qubos = cfg['problems']['scramble_qubos']
        self.problems = self._prep_problems()
        self.generalization_cfgs = self.cfg['problems'].get('generalization', {})

    def _prep_problems(self):
        ret = []
        for name in self.cfg['problems']['problems']:
            ret.append((PROBLEM_REGISTRY[name], self.cfg['problems'][name], name))
        return ret

    def gen_qubo_matrices(self, cls, n_problems, **kwargs):
        problems = cls.gen_problems(self.cfg, n_problems, **kwargs)
        qubo_matrices = [
            cls(self.cfg, **problem).gen_qubo_matrix()
            for problem in problems
        ]
        return problems, qubo_matrices

    def _gen_data(self, n_problems):
        qubo_size = self.qubo_size
        all_problems = []
        data = []
        # data = np.zeros(
        #     shape=(len(self.problems) * n_problems, qubo_size, qubo_size),
        #     dtype=np.float32
        # )

        generalized_params = self.cfg['problems'].get('generalization', {})
        generalized_order = 0
        if generalized_params:
            k = list(generalized_params.keys())[0]
            generalized_order = len(generalized_params[k])

        labels = []
        # labels = np.zeros(
        #     shape=(len(self.problems) * n_problems * (generalized_order + 1),),
        #     dtype=np.long
        # )
        for i, (cls, kwargs, name) in enumerate(self.problems):
            idx_start = i * n_problems
            idx_end = (i + 1) * n_problems
            problems, qubo_matrices = self.gen_qubo_matrices(
                cls, n_problems, **kwargs
            )
            generalized_params = self.generalization_cfgs.get(name, [])

            for params in generalized_params:
                args = kwargs.copy()
                args.update(params)
                problems_, qubo_matrices_ = self.gen_qubo_matrices(
                    cls, n_problems, **args
                )

                # Padding
                for k, matrix in enumerate(qubo_matrices_):
                    pad_width = qubo_size - matrix.shape[0]
                    matrix = np.pad(matrix, ((0, pad_width), (0, pad_width)), constant_values=0)
                    qubo_matrices_[k] = matrix
                problems.extend(problems_)
                qubo_matrices.extend(qubo_matrices_)

            all_problems.append(problems)
            if self.scramble_qubos:
                for j in range(len(qubo_matrices)):
                    if random.random() > .5:
                        rand_idx1 = random.randint(0, self.qubo_size - 1)
                        rand_idx2 = random.randint(0, self.qubo_size - 1)
                        val = qubo_matrices[j][[rand_idx2, rand_idx1]]
                        qubo_matrices[j][[rand_idx1, rand_idx2]] = val
            qubo_matrices = np.array(qubo_matrices)
            print(cls, qubo_matrices.shape)

            # Without normalization of some sort the NN won't learn.
            if not self.cfg["model"]["no_norm"]:
                if self.cfg["model"]["use_norm_divide"]:
                    qubo_matrices /= self.cfg["model"]["norm_multiply"]
                elif self.cfg["model"]["use_norm_multiply_input"]:
                    qubo_matrices /= np.max(np.abs(qubo_matrices))
                    qubo_matrices *= self.cfg["model"]["norm_multiply"]
                elif self.cfg["model"].get("norm_data_abs", False):
                    qubo_matrices =  qubo_matrices / np.max(np.abs(qubo_matrices))
                elif self.cfg["model"]["norm_data"]:
                    qubo_matrices /= np.max(np.abs(qubo_matrices))
                    qubo_matrices = (qubo_matrices + 1) / 2.
                else:
                    qubo_matrices = (
                        qubo_matrices - np.mean(qubo_matrices)
                    ) / np.std(qubo_matrices)

            # TODO !!
            # if qubo_matrices.shape[0] < n_problems:
            #     data[idx_start:idx_start + qubo_matrices.shape[0], :, :] = qubo_matrices
            # else:
            #     data[idx_start:idx_end, :, :] = qubo_matrices
            data.extend(qubo_matrices)
            # labels[idx_start:idx_end] = i
            labels.extend([i for _ in range(len(qubo_matrices))])

        print("TOTAL DATA LEN", len(data))
        data = np.array(data, dtype=np.float32)
        labels = np.array(labels, dtype=np.int32)
        return data, labels, all_problems

    def gen_data_gzip_pickle(self):
        data, labels, _ = self._gen_data(self.n_problems)
        fname = 'datasets/%s.pickle.gz' % self.cfg['cfg_id']
        with gzip.open(fname, 'wb+') as f:
            pickle.dump((data, labels), f)

    def gen_data_lmdb(self):
        data, labels, _ = self._gen_data(self.n_problems)
        dirpath = 'datasets/%s/'
        if self.cfg["use_big"]:
            dirpath = '/big/r/ratke/qubo_datasets/%s/'
        db = px.Writer(
            dirpath=dirpath % self.cfg['cfg_id'],
            map_size_limit=60000,
            ram_gb_limit=60
        )
        db.put_samples('input', data, 'target', labels)
        db.close()

    def gen_data_chunks(self):
        n_problems = self.n_problems
        for chunk in range(self.chunks):
            n_problems = self.n_problems // self.chunks
            data, labels, _ = self._gen_data(n_problems)

            fname = 'datasets/%s.%d.pickle.gz' % (self.cfg['cfg_id'], chunk)
            with gzip.open(fname, 'wb+') as f:
                pickle.dump((data, labels), f)

    def run_experiment(self, n_runs=1):
        lmdb_loader = LMDBDataLoader(self.cfg)
        for _ in range(n_runs):
            self.model_fname = self.get_model_fname()
            self.logger = Logger(self.model_fname, self.cfg)
            self.logger.log_config()
            optimizer = Optimizer(self.cfg, lmdb_loader, self.logger)
            optimizer.train()
            optimizer.save(self.model_fname)
            self._eval(optimizer)
            self.logger.close()

    def eval(self, model_fname):
        lmdb_loader = LMDBDataLoader(self.cfg)
        self.model_fname = self.get_model_fname()
        self.logger = Logger(self.model_fname, self.cfg)
        optimizer = Optimizer(self.cfg, lmdb_loader, self.logger)
        optimizer.load(model_fname)
        self._eval(optimizer)
        self.logger.close()

    def _eval(self, optimizer):
        misclassifications, _, _, mc_table = optimizer.eval()
        mc_prob = {
            self.cfg['problems']['problems'][int(k)]: v
            for k, v in misclassifications.items()
        }
        print(json.dumps(mc_prob, indent=4))
        self.logger.log_confusion_matrix(mc_table)

    def get_model_fname(self):
        rand_str = str(int(random.random() * 10e6))
        model_fname = "-".join([
            datetime.datetime.now().strftime("%y-%m-%d_%H:%M:%S"),
            rand_str,
            socket.gethostname(),
            self.cfg['cfg_id']
        ])
        return model_fname


class ReverseRegression(Classification):
    def __init__(self, cfg):
        super(ReverseRegression, self).__init__(cfg)

    def flatten_problem_parameters(self, all_problems):

        # TODO: Absolutely horrendous.

        gen_edges = self.cfg["model"]["gen_edges"]

        output_size = 0
        result = []
        for problem_class_params in all_problems:
            for problem in problem_class_params:
                problem = list(problem.values())
                curr_problem_result = []
                for part in problem:
                    if isinstance(part, nx.classes.graph.Graph):
                        if gen_edges:
                            A = np.asarray(sorted(part.edges))
                        else:
                            A = nx.to_numpy_matrix(part)
                        curr_problem_result.extend(list(A.flat))
                    elif isinstance(part, int) or isinstance(part, float):
                        curr_problem_result.append(part)
                    elif isinstance(part[0], list) or isinstance(part[0], tuple) or isinstance(part[0], np.ndarray):
                        if isinstance(part[0][0], tuple) or isinstance(part[0][0], list):
                            flat_list = [item for sublist in part for subsublist in sublist for item in subsublist]
                        else:
                            flat_list = [item for sublist in part for item in sublist]
                        flat_list = [float(x) for x in flat_list]
                        curr_problem_result.extend(flat_list)
                    else:
                        curr_problem_result.extend(part)
                result.append(curr_problem_result)
                output_size = max(output_size, len(curr_problem_result))
        return result, output_size

    def gen_apply_m2sat_customization(self, all_problems):
        """Apply a different labelling for M2SAT.

        Specifically, given the QUBO size is 64, we end up with a 64x64 label
        where all fields that have 1 denote a (T,T) clause, all fields that
        have (F,F) have 2, and so on (see code below).

        The original idea was to have the label consist of 4 of those 64x64
        matrices, but this ends up being very tough to train, with a ~16k
        dimensional output..

        Note that while I call this label, this is simply the NN output we
        train against.
        """
        for i, p in enumerate(all_problems[0]):
            clauses = p["clauses"]
            new_p = np.zeros(
                # shape=(4, self.qubo_size, self.qubo_size),
                shape=(1, self.qubo_size, self.qubo_size),
                dtype=np.float32
            )

            for clause in clauses:
                if clause[0][1] and clause[1][1]:
                    new_p[0][clause[0][0]][clause[1][0]] -= 2
                    new_p[0][clause[1][0]][clause[0][0]] -= 2
                if not clause[0][1] and not clause[1][1]:
                    new_p[0][clause[0][0]][clause[1][0]] += 1
                    new_p[0][clause[1][0]][clause[0][0]] += 1
                if clause[0][1] and not clause[1][1]:
                    new_p[0][clause[0][0]][clause[1][0]] -= 1
                    new_p[0][clause[1][0]][clause[0][0]] -= 1
                if not clause[0][1] and clause[1][1]:
                    new_p[0][clause[0][0]][clause[1][0]] += 2
                    new_p[0][clause[1][0]][clause[0][0]] += 2
            all_problems[0][i]["clauses"] = list(new_p.flat)

    def gen_apply_m3sat_customization(self, all_problems):
        """Apply a different labelling for M3SAT."""
        for i, p in enumerate(all_problems[0]):
            clauses = p["clauses"]
            new_p = np.zeros(
                shape=(1, self.qubo_size, self.qubo_size, self.qubo_size),
                dtype=np.float32
            )

            for clause in clauses:
                if clause[0][1] and clause[1][1] and clause[2][1]:
                    value = -4
                if clause[0][1] and not clause[1][1] and clause[2][1]:
                    value = -3
                if clause[0][1] and clause[1][1] and not clause[2][1]:
                    value = -2
                if clause[0][1] and not clause[1][1] and not clause[2][1]:
                    value = -1
                if not clause[0][1] and clause[1][1] and clause[2][1]:
                    value = 1
                if not clause[0][1] and not clause[1][1] and clause[2][1]:
                    value = 2
                if not clause[0][1] and not clause[1][1] and not clause[2][1]:
                    value = 3
                if not clause[0][1] and clause[1][1] and not clause[2][1]:
                    value = 4

                for permutation in itertools.permutations(clause, 3):
                    x = permutation[0][0]
                    y = permutation[1][0]
                    z = permutation[2][0]
                    new_p[0][x][y][z] += value
            all_problems[0][i]["clauses"] = list(new_p.flat)

    def gen_data_lmdb(self):
        data, labels, all_problems = self._gen_data(self.n_problems)

        if self.cfg['problems']['problems'] == ["M2SAT"]:
            self.gen_apply_m2sat_customization(all_problems)
            print(data[0])
            print(all_problems[0][0]["clauses"])

        if self.cfg['problems']['problems'] == ["M3SAT"]:
            len_ = len(all_problems[0])
            data = data[:len_]
            labels = labels[:len_]

            self.gen_apply_m3sat_customization(all_problems)
            print(data[0])
            print(all_problems[0][0]["clauses"])

        if self.cfg['problems']['problems'] == ["QA"]:
            print("Simplifying input space..")

            new_data = []

            size = self.cfg["problems"]["QA"]["size"]
            max_full_size = None

            generalized_params = self.cfg['problems'].get('generalization', {}).get('QA', [])
            for m in range(len(generalized_params) + 1):
                if m > 0:
                    size = self.cfg["problems"]["generalization"][m - 1]["size"]
                qubo_size = int(size ** 2)
                for n in range(self.n_problems):
                    d = data[self.n_problems * m + n]
                    d = d[:qubo_size, :qubo_size]
                    tmp_data = []
                    triu_idx = np.triu_indices(size, 1)
                    for i, j in zip(*triu_idx):
                        quadrant = d[i*size:(i+1)*size, j*size:(j+1)*size]
                        x = quadrant[np.triu_indices(size, 1)]
                        tmp_data.extend(x)

                    if m > 0:
                        pad_width = max_full_size - len(tmp_data)
                        tmp_data = np.pad(tmp_data, (0, pad_width))

                    new_data.append(tmp_data)
                if m == 0:
                    max_full_size = len(new_data[0])
            data = np.array(new_data)

            rm_perc = self.cfg["problems"]["QA"].get("remove_percentile_smallest", None)
            if rm_perc is not None:
                for d in data:
                    perc = np.percentile(d, rm_perc)
                    d[np.where(d < perc)] = 0.

            print(data[0])
            print(all_problems[0][0])

        if self.cfg['problems']['problems'] == ["SP"]:
            len_ = len(all_problems[0])
            data = data[:len_]
            labels = labels[:len_]

            print(data[0])
            print(all_problems[0][0])

        if self.cfg['problems']['problems'] == ["QK"]:
            print(data[0])
            print(all_problems[0][0])

            special_norm = self.cfg['problems']['QK'].get('special_norm', False)
            # Assumption: We know where the budgets lie.
            budgets_idx = self.cfg['problems']['QK']['size']
            multiplier = 2 ** (self.qubo_size - budgets_idx)

            print("QK SPEC NORM?", special_norm, budgets_idx, multiplier)

            if special_norm:
                new_data = []
                for k, d in enumerate(data):
                    d_new = d[:budgets_idx, :budgets_idx]
                    y = d[:budgets_idx, budgets_idx].reshape(1, -1)

                    # np.fill_diagonal(d_new, np.diag(d_new) + y * 2 * 6)
                    np.fill_diagonal(d_new, np.diag(d_new) + y * 2 * multiplier)

                    d_new += np.amax([d_new, y.T @ y], axis=0)

                    x = np.concatenate((d_new.flatten(), (-y * 2).flatten(), (y.T @ y).flatten() / 100.))
                    new_data.append(x)
                data = np.array(new_data)
                print(data[0])

        if self.cfg['problems']['problems'] == ["MCQ"]:
            len_ = len(all_problems[0])
            data = data[:len_]
            labels = labels[:len_]

            data -= 2  # NOTE: LOL

            print(data[0])
            print(all_problems[0][0])

        all_problems_flat, output_size = self.flatten_problem_parameters(all_problems)

        prob_type = self.cfg['problems']['problems']
        if prob_type == ["MVC"] or prob_type == ["MC"]:
            for i, prob in enumerate(all_problems_flat):
                len_ = int(len(prob) ** .5)
                pad_width = int(output_size ** .5) - len_
                prob = np.reshape(prob, (len_, len_))
                all_problems_flat[i] = np.pad(
                    prob,
                    ((0, pad_width), (0, pad_width)),
                    'constant',
                    constant_values=0
                ).astype(float).reshape(output_size)
        elif prob_type == ["GC"]:
            for i, prob in enumerate(all_problems_flat):
                len_ = int(len(prob) ** .5)
                pad_width = int(output_size ** .5) - len_
                val = prob[-1]
                prob = np.reshape(prob[:-1], (len_, len_))
                all_problems_flat[i] = np.pad(
                    prob,
                    ((0, pad_width), (0, pad_width)),
                    'constant',
                    constant_values=0
                ).astype(float).reshape(output_size - 1)
                all_problems_flat[i] = np.append(all_problems_flat[i], val)
        else:
            for i, prob in enumerate(all_problems_flat):
                all_problems_flat[i] = np.pad(
                    prob,
                    (0, output_size - len(prob)),
                    'constant',
                    constant_values=(0, 0)
                ).astype(float)

        all_problems_flat = np.array(all_problems_flat, dtype=float)

        print(data.shape, labels.shape, all_problems_flat.shape)

        print("NORM?", not self.cfg["model"]["no_norm"], self.cfg["model"]["norm_multiply"])

        # NOTE: We are using min max normalization here.. Not standardization
        # like with classification.
        if not self.cfg["model"]["no_norm"]:
            print("use norm multiply?", self.cfg["model"]["use_norm_multiply"])
            if self.cfg["model"]["use_norm_divide"]:
                all_problems_flat /= self.cfg["model"]["norm_multiply"]
            elif self.cfg["model"]["use_norm_multiply"]:
                all_problems_flat /= np.max(np.abs(all_problems_flat))
                all_problems_flat *= self.cfg["model"]["norm_multiply"]
                # qa4_diffnorm used 5., qa4_diffnorm_v2 used 10.
            elif self.cfg["model"].get("norm_data_abs", False):
                all_problems_flat /= np.max(np.abs(all_problems_flat))
            else:
                all_problems_flat /= np.max(np.abs(all_problems_flat))
                all_problems_flat = (all_problems_flat + 1) / 2.

        # all_problems_flat = (
        #     all_problems_flat - np.mean(all_problems_flat)
        # ) / np.std(all_problems_flat)

        db = px.Writer(
            dirpath='datasets/%s/' % self.cfg['cfg_id'],
            map_size_limit=60000,
            ram_gb_limit=60
        )
        db.put_samples('input', data, 'target', labels, 'prob', all_problems_flat)
        db.close()

        with open('datasets/%s/cfg.pickle' % self.cfg['cfg_id'], 'wb+') as f:
            pickle.dump(output_size, f)

    def run_experiment(self, n_runs=1):
        part = self.cfg["model"].get("part", None)
        if part:
            lmdb_loader = LMDBDataLoader(self.cfg, reverse=True, part=part)
        else:
            lmdb_loader = LMDBDataLoader(self.cfg, reverse=True)

        dirpath = 'datasets/'
        if self.cfg["use_big"]:
            dirpath = '/big/r/ratke/qubo_datasets/'
        with open(dirpath + '%s/cfg.pickle' % self.cfg['dataset_id'], 'rb') as f:
            output_size = pickle.load(f)

        for _ in range(n_runs):
            self.model_fname = self.get_model_fname()
            self.logger = Logger(self.model_fname, self.cfg)
            self.logger.log_config()
            optimizer = ReverseOptimizer(self.cfg, lmdb_loader, self.logger, output_size)
            # optimizer = RNNOptimizer(self.cfg, lmdb_loader, self.logger, output_size)
            optimizer.train()
            optimizer.save(self.model_fname)
            self.logger.close()

    def eval(self, model_fname):
        part = self.cfg["model"].get("part", None)
        if part:
            lmdb_loader = LMDBDataLoader(self.cfg, reverse=True, part=part)
        else:
            lmdb_loader = LMDBDataLoader(self.cfg, reverse=True)

        with open('datasets/%s/cfg.pickle' % self.cfg['dataset_id'], 'rb') as f:
            output_size = pickle.load(f)

        self.model_fname = self.get_model_fname()
        self.logger = Logger(self.model_fname, self.cfg)
        optimizer = ReverseOptimizer(self.cfg, lmdb_loader, self.logger, output_size)
        optimizer.load(model_fname, output_size)
        self._eval(optimizer)
        self.logger.close()

    def _eval(self, optimizer):
        avg_loss, problem_losses, tot3_fp, tot3_fn, tot3 = optimizer.eval(0, debug=True)
        print(avg_loss, problem_losses)
        print(tot3_fp, tot3_fn, tot3)
        print((tot3_fp + tot3_fn) / tot3)


class AutoEncoder(Classification):
    def auto_encoder_prototype(self):
        # TODO Refactor this into its own class.

        # Prepare data.
        lmdb_loader = LMDBDataLoader(self.cfg)

        # TODO Support n_runs later.
        # for _ in range(n_runs):

        self.model_fname = self.get_model_fname()
        self.logger = Logger(self.model_fname, self.cfg)
        self.logger.log_config()

        optimizer = AutoEncoderOptimizer(self.cfg, lmdb_loader, self.logger)
        optimizer.train()
        optimizer.save(self.model_fname)

        self.logger.close()


class DefeatClassification(Classification):
    def __init__(self, cfg):
        super(DefeatClassification, self).__init__(cfg)

    def _gen_data(self, n_problems):
        print("Generating data to defeat classification.")

        qubo_size = self.qubo_size

        all_problems = []
        data = []
        labels = []

        generalized_params = self.cfg['problems'].get('generalization', {})
        generalized_order = 0
        if generalized_params:
            k = list(generalized_params.keys())[0]
            generalized_order = len(generalized_params[k])

        for i, (cls, kwargs, name) in enumerate(self.problems):
            print(name)
            idx_start = i * n_problems
            idx_end = (i + 1) * n_problems
            problems, qubo_matrices = self.gen_qubo_matrices(
                cls, n_problems, **kwargs
            )

            generalized_params = self.generalization_cfgs.get(name, [])

            for params in generalized_params:
                args = kwargs.copy()
                args.update(params)
                problems_, qubo_matrices_ = self.gen_qubo_matrices(
                    cls, n_problems, **args
                )

                # Padding
                for k, matrix in enumerate(qubo_matrices_):
                    pad_width = qubo_size - matrix.shape[0]
                    matrix = np.pad(matrix, ((0, pad_width), (0, pad_width)), constant_values=0)
                    qubo_matrices_[k] = matrix
                problems.extend(problems_)
                qubo_matrices.extend(qubo_matrices_)

            do_add_noise = name == "MC" or name == "MVC"
            # do_add_noise = False

            noise_divisor = self.cfg['model']['noise_divisor']

            all_problems.append(problems)
            for j in range(len(qubo_matrices)):
                if self.scramble_qubos:
                    if random.random() > .5:
                        rand_idx1 = random.randint(0, self.qubo_size - 1)
                        rand_idx2 = random.randint(0, self.qubo_size - 1)
                        val = qubo_matrices[j][[rand_idx2, rand_idx1]]
                        qubo_matrices[j][[rand_idx1, rand_idx2]] = val

                if name == "MC":
                    qubo_matrices[j] *= -1
                    # qubo_matrices[j] /= 2.

                if name == "MVC":
                    qubo_matrices[j] /= 8.

                if do_add_noise:
                    noise = np.random.random(size=qubo_matrices[j].shape) / noise_divisor
                    np.fill_diagonal(noise, 0.)
                    qubo_matrices[j] += noise

                # Normalize.
                qubo_matrices[j] /= abs(qubo_matrices[j]).max()

                if name == "MVC" or name == "MC":
                    # qubo_matrices[j][np.where(qubo_matrices[j] > 0.05)] = 0.03
                    qubo_matrices[j][np.where(qubo_matrices[j] > 0.03)] = 0.03
                if name == "GC" or name == "TSP":
                    diag = np.diag(qubo_matrices[j])
                    qubo_matrices[j][np.where(qubo_matrices[j] > 0.03)] = 0.03
                    np.fill_diagonal(qubo_matrices[j], diag / 2)
            qubo_matrices = np.array(qubo_matrices)
            print(cls, qubo_matrices.shape)

            data.extend(qubo_matrices)
            labels.extend([i for _ in range(len(qubo_matrices))])

        print("TOTAL DATA LEN", len(data))
        data = np.array(data, dtype=np.float32)
        labels = np.array(labels, dtype=np.int32)
        return data, labels, all_problems


class A3(Classification):
    def __init__(self, cfg):
        super(A3, self).__init__(cfg)

    def solve_qubo2(self, item):
        Q = qubovert.utils.matrix_to_qubo(item.reshape(self.qubo_size, self.qubo_size))
        sol = Q.solve_bruteforce(all_solutions=False)
        sol_ = [0 for _ in range(self.qubo_size)]
        for k, v in sol.items():
            sol_[k] = v
        return sol_

    def _gen_data(self, n_problems):
        print("Generating data for A3.")

        qubo_size = self.qubo_size

        all_problems = []
        data = []
        labels = []

        generalized_params = self.cfg['problems'].get('generalization', {})
        generalized_order = 0
        if generalized_params:
            k = list(generalized_params.keys())[0]
            generalized_order = len(generalized_params[k])

        for i, (cls, kwargs, name) in enumerate(self.problems):
            print(name)
            idx_start = i * n_problems
            idx_end = (i + 1) * n_problems
            problems, qubo_matrices = self.gen_qubo_matrices(
                cls, n_problems, **kwargs
            )

            generalized_params = self.generalization_cfgs.get(name, [])

            for params in generalized_params:
                args = kwargs.copy()
                args.update(params)
                problems_, qubo_matrices_ = self.gen_qubo_matrices(
                    cls, n_problems, **args
                )

                # Padding
                for k, matrix in enumerate(qubo_matrices_):
                    pad_width = qubo_size - matrix.shape[0]
                    matrix = np.pad(matrix, ((0, pad_width), (0, pad_width)), constant_values=0)
                    qubo_matrices_[k] = matrix
                problems.extend(problems_)
                qubo_matrices.extend(qubo_matrices_)

            # do_add_noise = name == "MC" or name == "MVC"

            noise_divisor = self.cfg['model']['noise_divisor']

            all_problems.append(problems)
            for j in range(len(qubo_matrices)):
                if self.scramble_qubos:
                    if random.random() > .5:
                        rand_idx1 = random.randint(0, self.qubo_size - 1)
                        rand_idx2 = random.randint(0, self.qubo_size - 1)
                        val = qubo_matrices[j][[rand_idx2, rand_idx1]]
                        qubo_matrices[j][[rand_idx1, rand_idx2]] = val

                if name == "MC":
                    qubo_matrices[j] *= -1
                    # qubo_matrices[j] /= 2.

                # if name == "MVC":
                #     qubo_matrices[j] /= 8.

                # if do_add_noise:
                #     noise = np.random.random(size=qubo_matrices[j].shape) / noise_divisor
                #     np.fill_diagonal(noise, 0.)
                #     qubo_matrices[j] += noise

                # Normalize.
                # qubo_matrices[j] /= abs(qubo_matrices[j]).max()
                qubo_matrices[j] = qubo_matrices[j] / abs(qubo_matrices[j]).max()

                # if name == "MVC" or name == "MC":
                #     # qubo_matrices[j][np.where(qubo_matrices[j] > 0.05)] = 0.03
                #     qubo_matrices[j][np.where(qubo_matrices[j] > 0.03)] = 0.03
                # if name == "GC" or name == "TSP":
                #     diag = np.diag(qubo_matrices[j])
                #     qubo_matrices[j][np.where(qubo_matrices[j] > 0.03)] = 0.03
                #     np.fill_diagonal(qubo_matrices[j], diag / 2)
            qubo_matrices = np.array(qubo_matrices)
            print(cls, qubo_matrices.shape)

            data.append(qubo_matrices)  # NOTE: We append, not extend here!!
            labels.append([self.solve_qubo2(q) for q in qubo_matrices])
            # labels.append([i for _ in range(len(qubo_matrices))])

        print("TOTAL DATA LEN", len(data))
        data = np.array(data, dtype=np.float32)
        labels = np.array(labels, dtype=np.int32)
        return data, labels, all_problems

    def run_experiment(self, n_runs=1):
        print("Running A3.")
        print(torch.get_num_threads())
        lmdb_loader = LMDBDataLoader(self.cfg)
        for _ in range(n_runs):
            self.model_fname = self.get_model_fname()
            self.logger = Logger(self.model_fname, self.cfg)
            self.logger.log_config()
            optimizer = A3Optimizer(self.cfg, lmdb_loader, self.logger)
            optimizer.train()
            optimizer.save(self.model_fname)
            self._eval(optimizer)
            self.logger.close()

    def _eval(self, optimizer):
        # TODO
        pass


class R1(Classification):
    def __init__(self, cfg):
        super(R1, self).__init__(cfg)

    def _gen_data(self, n_problems):
        print("Generating data for R1.")

        qubo_size = self.qubo_size

        all_problems = []
        data = []
        labels = []

        generalized_params = self.cfg['problems'].get('generalization', {})
        generalized_order = 0
        if generalized_params:
            k = list(generalized_params.keys())[0]
            generalized_order = len(generalized_params[k])

        for i, (cls, kwargs, name) in enumerate(self.problems):
            print(name)
            idx_start = i * n_problems
            idx_end = (i + 1) * n_problems
            problems, qubo_matrices = self.gen_qubo_matrices(
                cls, n_problems, **kwargs
            )

            generalized_params = self.generalization_cfgs.get(name, [])

            for params in generalized_params:
                args = kwargs.copy()
                args.update(params)
                problems_, qubo_matrices_ = self.gen_qubo_matrices(
                    cls, n_problems, **args
                )

                # Padding
                for k, matrix in enumerate(qubo_matrices_):
                    pad_width = qubo_size - matrix.shape[0]
                    matrix = np.pad(matrix, ((0, pad_width), (0, pad_width)), constant_values=0)
                    qubo_matrices_[k] = matrix
                problems.extend(problems_)
                qubo_matrices.extend(qubo_matrices_)

            all_problems.append(problems)
            for j in range(len(qubo_matrices)):
                if self.scramble_qubos:
                    if random.random() > .5:
                        rand_idx1 = random.randint(0, self.qubo_size - 1)
                        rand_idx2 = random.randint(0, self.qubo_size - 1)
                        val = qubo_matrices[j][[rand_idx2, rand_idx1]]
                        qubo_matrices[j][[rand_idx1, rand_idx2]] = val

                if name == "MC":
                    qubo_matrices[j] *= -1

                # Normalize.
                qubo_matrices[j] /= abs(qubo_matrices[j]).max()

            qubo_matrices = np.array(qubo_matrices)
            print(cls, qubo_matrices.shape)

            data.extend(qubo_matrices)
            labels.extend([i for _ in range(len(qubo_matrices))])

        print("TOTAL DATA LEN", len(data))
        data = np.array(data, dtype=np.float32)
        labels = np.array(labels, dtype=np.int32)
        return data, labels, all_problems

    def run_experiment(self, n_runs=1):
        print("Running R1.")
        lmdb_loader = LMDBDataLoader(self.cfg)
        for _ in range(n_runs):
            self.model_fname = self.get_model_fname()
            self.logger = Logger(self.model_fname, self.cfg)
            self.logger.log_config()
            optimizer = Resistance1(self.cfg, lmdb_loader, self.logger)
            optimizer.train()
            optimizer.save(self.model_fname)
            self._eval(optimizer)
            self.logger.close()

    def _eval(self, optimizer):
        # TODO
        pass


class R2(Classification):
    def __init__(self, cfg):
        super(R2, self).__init__(cfg)

    def _gen_data(self, n_problems):
        print("Generating data for R2.")

        qubo_size = self.qubo_size

        all_problems = []
        data = []
        labels = []

        generalized_params = self.cfg['problems'].get('generalization', {})
        generalized_order = 0
        if generalized_params:
            k = list(generalized_params.keys())[0]
            generalized_order = len(generalized_params[k])

        for i, (cls, kwargs, name) in enumerate(self.problems):
            print(name)
            idx_start = i * n_problems
            idx_end = (i + 1) * n_problems
            problems, qubo_matrices = self.gen_qubo_matrices(
                cls, n_problems, **kwargs
            )

            generalized_params = self.generalization_cfgs.get(name, [])

            for params in generalized_params:
                args = kwargs.copy()
                args.update(params)
                problems_, qubo_matrices_ = self.gen_qubo_matrices(
                    cls, n_problems, **args
                )

                # Padding
                for k, matrix in enumerate(qubo_matrices_):
                    pad_width = qubo_size - matrix.shape[0]
                    matrix = np.pad(matrix, ((0, pad_width), (0, pad_width)), constant_values=0)
                    qubo_matrices_[k] = matrix
                problems.extend(problems_)
                qubo_matrices.extend(qubo_matrices_)

            all_problems.append(problems)
            for j in range(len(qubo_matrices)):
                if self.scramble_qubos:
                    if random.random() > .5:
                        rand_idx1 = random.randint(0, self.qubo_size - 1)
                        rand_idx2 = random.randint(0, self.qubo_size - 1)
                        val = qubo_matrices[j][[rand_idx2, rand_idx1]]
                        qubo_matrices[j][[rand_idx1, rand_idx2]] = val

                if name == "MC":
                    qubo_matrices[j] *= -1

                # Normalize.
                qubo_matrices[j] /= abs(qubo_matrices[j]).max()

            qubo_matrices = np.array(qubo_matrices)
            print(cls, qubo_matrices.shape)

            data.extend(qubo_matrices)
            labels.extend([i for _ in range(len(qubo_matrices))])

        print("TOTAL DATA LEN", len(data))
        data = np.array(data, dtype=np.float32)
        labels = np.array(labels, dtype=np.int32)
        return data, labels, all_problems

    def run_experiment(self, n_runs=1):
        print("Running R2.")
        lmdb_loader = LMDBDataLoader(self.cfg)
        for _ in range(n_runs):
            self.model_fname = self.get_model_fname()
            self.logger = Logger(self.model_fname, self.cfg)
            self.logger.log_config()
            optimizer = Resistance2(self.cfg, lmdb_loader, self.logger)
            optimizer.train()
            optimizer.save(self.model_fname)
            self._eval(optimizer)
            self.logger.close()

    def _eval(self, optimizer):
        # TODO
        pass


class QbsolvRegression(Classification):
    def __init__(self, cfg):
        super(QbsolvRegression, self).__init__(cfg)
        self.qubo_size = self.cfg["problems"]["qubo_size"]

    def gen_qubo_matrices(self, cls, n_problems, **kwargs):
        problems = cls.gen_problems(self.cfg, n_problems, **kwargs)
        qubo_matrices = []
        for i, problem in enumerate(problems):
            if i % 100000 == 0:
                print(i)
            qubo_matrices.append(cls(self.cfg, **problem).gen_qubo_matrix())
            del problem
        return qubo_matrices

    def _gen_data(self, n_problems):
        qubo_size = self.qubo_size
        data = []

        labels = []
        for i, (cls, kwargs, name) in enumerate(self.problems):
            cls = qubo_nn.problems.max_cut.MaxCutMemoryEfficient
            qubo_matrices = self.gen_qubo_matrices(
                cls, n_problems, **kwargs
            )

            qubo_matrices = np.array(qubo_matrices)
            print(cls, qubo_matrices.shape)

            data.extend(qubo_matrices)
            labels.extend([i for _ in range(len(qubo_matrices))])

        print("TOTAL DATA LEN", len(data))
        data = np.array(data, dtype=np.dtype('b'))
        labels = np.array(labels, dtype=np.int32)
        return data, labels, []

    def solve_qubo(self, qubo):
        qubo = qubo.copy().astype(np.dtype('i4'))
        Q = qubovert.utils.matrix_to_qubo(qubo)
        sol = Q.solve_bruteforce(all_solutions=False)
        sol_ = [0 for _ in range(self.qubo_size)]
        for k, v in sol.items():
            sol_[k] = v
        # return torch.FloatTensor(sol_)
        return sol_

    def gen_data_lmdb(self):
        data, _, _ = self._gen_data(self.n_problems)

        labels = []

        for i, qubo in enumerate(data):
            if i % 100000 == 0:
                print(i)
            labels.append(self.solve_qubo(qubo))
        labels = np.array(labels)

        print(data.shape, labels.shape)

        print("NORM?", not self.cfg["model"]["no_norm"], self.cfg["model"]["norm_multiply"])

        db = px.Writer(
            dirpath='datasets/%s/' % self.cfg['cfg_id'],
            map_size_limit=60000,
            ram_gb_limit=60
        )
        db.put_samples('input', data, 'target', labels)
        db.close()

    def run_experiment(self, n_runs=1):
        lmdb_loader = LMDBDataLoader(self.cfg)

        for _ in range(n_runs):
            self.model_fname = self.get_model_fname()
            self.logger = Logger(self.model_fname, self.cfg)
            self.logger.log_config()
            optimizer = QbsolvOptimizer(self.cfg, lmdb_loader, self.logger)
            optimizer.train()
            optimizer.save(self.model_fname)
            self.logger.close()


class RedAERegression(Classification):
    def __init__(self, cfg):
        super(RedAERegression, self).__init__(cfg)
        self.qubo_size = self.cfg["problems"]["qubo_size"]

    def gen_data_lmdb(self):
        data, labels, _ = self._gen_data(self.n_problems)

        # TODO: NP is weird tbh fam. Get rid of this.
        # if self.cfg['problems']['problems'] == ["NP"]:
        #     for i, d in enumerate(data):
        #         np.fill_diagonal(data[i], np.diag(data[i] / 4.))
        # print(data[0])

        print(data.shape, labels.shape)

        db = px.Writer(
            dirpath='datasets/%s/' % self.cfg['cfg_id'],
            map_size_limit=60000,
            ram_gb_limit=60
        )
        db.put_samples('input', data, 'target', labels)
        db.close()

    def run_experiment(self, n_runs=1):
        lmdb_loader = LMDBDataLoader(self.cfg)

        for _ in range(n_runs):
            self.model_fname = self.get_model_fname()
            self.logger = Logger(self.model_fname, self.cfg)
            self.logger.log_config()
            optimizer = RedAEOptimizer(self.cfg, lmdb_loader, self.logger)
            optimizer.train()
            optimizer.save(self.model_fname)
            self.logger.close()



import copy
import json


problems = ["NP", "MC", "MVC", "SP", "M2SAT", "SPP", "GC", "QA", "QK", "M3SAT", "TSP", "GI", "SGI", "MCQ"]


cfg = {
    "red2": {
        "dataset_id": "red2",
        "base_cfg": "red_base",
        "problems": {
            "n_problems": 100000,
            "problems": ["NP"]
        },
        "model": {
            # "n_epochs": 100,
            "lr": 10.0,
            "fc_sizes": [[4096], [4096]]
        }
    }
}


with open("simulations.json", "a+") as f:
    for problem in problems:
        dataset_cfg_id = "red_" + problem.lower() + "_1"
        for n in range(1, 20):
            size = 4096 * (n / 20)
            cfg_ = copy.deepcopy(cfg)
            cfg_["red2"]["model"]["fc_sizes"][0] = [int(size)]
            cfg_["red2"]["problems"]["problems"] = [problem]
            cfg_id = "red_" + problem.lower() + "_" + str(n)
            cfg_[cfg_id] = cfg_["red2"]
            cfg_[cfg_id]["dataset_id"] = dataset_cfg_id
            del cfg_["red2"]
            f.write("    " + json.dumps(cfg_)[1:-1] + ",\n")



import sys
import qubovert
import pyxis as px
import numpy as np
from sklearn.metrics.pairwise import cosine_distances

from qubo_nn.data import LMDBDataLoader


QUBO_SIZE = 9


def solve_qubo2(item):
    Q = qubovert.utils.matrix_to_qubo(item.reshape(QUBO_SIZE, QUBO_SIZE))
    sol = Q.solve_bruteforce(all_solutions=False)
    sol_ = [0 for _ in range(QUBO_SIZE)]
    for k, v in sol.items():
        sol_[k] = v
    return sol_


cfg = {
    'dataset_id': 'a3dbg_all',
    'use_big': False,
    'model': {
        'batch_size': 1,
        'train_eval_split': 1.0,
        'shuffle_data': False
    }
}

lmdb_loader = LMDBDataLoader(cfg)
loader = lmdb_loader.train_data_loader
data = list(loader)

labels = [d[1][0].numpy() for d in data]
data = [d[0][0].numpy() for d in data]
data_copy = [d.copy() for d in data]

print("Solving..")
sols = [np.array([solve_qubo2(d) for d in curr_data]) for curr_data in data]
prob = ["NP", "MC", "MVC", "SP", "M2SAT", "SPP", "QA", "QK", "M3SAT", "TSP", "MCQ"]
# prob = ["NP", "MVC", "M2SAT", "M3SAT", "QA", "SP", "SPP", "QK", "MC", "TSP", "MCQ"]
pairs = [
    (0, 2),
    (4, 8),
    (6, 3),
    (5, 7),
    (1, 9),
    (10, 9)
]


def make_similar(sols_all, factors):
    for i, j in pairs:
        if i == 10:
            diff2 = (data_copy[j] - data[i]) * factors[i]
            data[i] += diff1
            check(i, data[i], data_copy[i], sols_all[i])  # i
        else:
            diff1 = (data_copy[j] - data[i]) * factors[i]
            diff2 = (data_copy[i] - data[j]) * factors[j]
            data[i] += diff1
            data[j] += diff2
            check(i, data[i], data_copy[i], sols_all[i])  # i
            check(j, data[j], data_copy[j], sols_all[j])  # j


def check(i, our_data, cached_data, sols):
    total_len = QUBO_SIZE * len(our_data)
    new_sol = np.array([solve_qubo2(d) for d in our_data])
    # print(abs(new_sol - sols).sum())
    binary_err = 0
    real_err = 0
    total_val = 0
    for k, (new_sol_, sol) in enumerate(zip(new_sol, sols)):
        best_val = sol.T @ cached_data[k] @ sol
        new_val = new_sol_.T @ cached_data[k] @ new_sol_

        binary_err += min(
            abs(new_sol_ - sol).sum(),
            abs((1 - new_sol_) - sol).sum()
        )
        real_err += abs(best_val - new_val)
        total_val += abs(best_val)
    # print(prob[i], binary_err, binary_err / total_len, real_err, real_err / total_val)
    print(prob[i], "\t", binary_err / total_len)
    return our_data

factors_dict = {
    "0": [.5, .5, .5, .5, .5, .5, .5, .5, .5, .5, .5],
    "1": [.5, .4, .5, .4, .5, .4, .5, .4, .5, .4, .5],
    "2": [.4, .4, .4, .4, .4, .4, .4, .4, .4, .4, .4],
    "3": [.4, .3, .4, .3, .4, .3, .4, .3, .4, .3, .4],
    "4": [.3, .3, .3, .3, .3, .3, .3, .3, .3, .3, .3],
    "5": [.2, .2, .2, .2, .2, .2, .2, .2, .2, .2, .2],
    "6": [.1, .1, .1, .1, .1, .1, .1, .1, .1, .1, .1],
    "7": [0.05] * 11,
    "8": [0.] * 11
}
factors = factors_dict[sys.argv[1]]
print("Using factors", factors)
# factors = [.5, .5, .5, .5, .5, .5, .5, .5, .5, .5, .5]
# factors = [.5, .4, .5, .4, .5, .4, .5, .4, .5, .4, .5]
# factors = [.4, .4, .4, .4, .4, .4, .4, .4, .4, .4, .4]
# factors = [.4, .3, .4, .3, .4, .3, .4, .3, .4, .3, .4]
# factors = [.3, .3, .3, .3, .3, .3, .3, .3, .3, .3, .3]
# factors = [.2, .2, .2, .2, .2, .2, .2, .2, .2, .2, .2]
# factors = [.1, .1, .1, .1, .1, .1, .1, .1, .1, .1, .1]
# factors = [0.05] * 11
# factors = [0.] * 11
make_similar(sols, factors)

data_flattened = [d.flatten() for d in data]
print("COS DIST")
print(cosine_distances(data_flattened).round(2))


def save(data, labels, cfg_id):
    data_new = []
    for d in data:
        data_new.extend(d)
    data = np.array(data_new)
    labels_new = []
    for label in labels:
        labels_new.extend(label)
    labels = np.array(labels_new)
    dirpath = 'datasets/%s/'
    db = px.Writer(
        dirpath=dirpath % cfg_id,
        map_size_limit=60000,
        ram_gb_limit=60
    )
    db.put_samples('input', data, 'target', labels)
    db.close()


cfg_id = "sim_pair2_%d_%d" % tuple(int(100 * f)  for f in factors[:2])
save(data, labels, cfg_id)
print(cfg_id)



import io
import json
import pickle

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib as mpl

import PIL.Image
from torchvision.transforms import ToTensor
from torch.utils.tensorboard import SummaryWriter


mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')


def truncate_colormap(cmapIn='jet', minval=0.0, maxval=1.0, n=100):
    cmapIn = plt.get_cmap(cmapIn)

    new_cmap = colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmapIn.name, a=minval, b=maxval),
        cmapIn(np.linspace(minval, maxval, n)))

    return new_cmap


class Logger:
    def __init__(self, model_fname, cfg):
        self.cfg = cfg
        self.model_fname = model_fname
        self.writer = SummaryWriter(log_dir='runs/' + model_fname)

    def log_config(self):
        self.writer.add_text('Info/Config', json.dumps(self.cfg), 0)

    def log_train(self, data, n_iter):
        self.writer.add_scalar('Loss/Train', data['loss_train'], n_iter)

    def log_eval(self, data, n_iter):
        self.writer.add_scalar('Loss/Eval', data['loss_eval'], n_iter)
        for k, v in data['Problem_Misclassifications'].items():
            self.writer.add_scalar(
                'Problem_Misclassifications/' + k,
                v,
                n_iter
            )
        self.writer.add_scalar(
            'Total_Misclassifications',
            data['Total_Misclassifications'],
            n_iter
        )

    def log_eval_reverse(self, data, n_iter):
        self.writer.add_scalar('Loss/Eval', data['loss_eval'], n_iter)

    def log_custom_reverse_kpi(self, kpi, data, n_iter):
        self.writer.add_scalar('Custom/' + kpi, data, n_iter)

    def log_confusion_matrix(self, mc_table):
        cmap_mod = truncate_colormap('Greens', minval=.4, maxval=.99)
        fig, ax = plt.subplots(figsize=(10, 10))
        im = ax.imshow(mc_table, cmap=cmap_mod, vmin=0, vmax=1)

        plt.ylabel('True label')
        plt.xlabel('Predicted label')
        plt.tight_layout()

        # Text annotations
        for i in range(mc_table.shape[0]):
            for j in range(mc_table.shape[0]):
                ax.text(j, i, '%.02f' % mc_table[i][j], ha="center", va="center", color="w")

        buf = io.BytesIO()
        plt.savefig(buf, format='jpeg')

        problems = self.cfg['problems']['problems']
        plt.xticks(list(range(len(problems))), problems)
        plt.yticks(list(range(len(problems))), problems)

        plt.close(fig)
        buf.seek(0)
        image = PIL.Image.open(buf)
        image = ToTensor()(image).unsqueeze(0)
        image = image.reshape((3, 1000, 1000))

        self.writer.add_image('Info/Confusion_matrix', image, 0)

        with open('runs/' + self.model_fname + '/confusion_matrix_data.pickle', 'wb+') as f:
            pickle.dump(mc_table, f)

    def close(self):
        self.writer.close()



import qubovert
import pyxis as px
import numpy as np
from scipy.spatial import distance

from qubo_nn.data import LMDBDataLoader


QUBO_SIZE = 8


def solve_qubo2(item):
    Q = qubovert.utils.matrix_to_qubo(item.reshape(QUBO_SIZE, QUBO_SIZE))
    sol = Q.solve_bruteforce(all_solutions=False)
    sol_ = [0 for _ in range(QUBO_SIZE)]
    for k, v in sol.items():
        sol_[k] = v
    return sol_


cfg = {
    'dataset_id': 'a3dbg',
    'use_big': False,
    'model': {
        'batch_size': 1,
        'train_eval_split': 1.0,
        'shuffle_data': False
    }
}

lmdb_loader = LMDBDataLoader(cfg)
loader = lmdb_loader.train_data_loader
data = list(loader)

np_data = data[0][0][0].numpy()
mc_data = data[1][0][0].numpy()
mvc_data = data[2][0][0].numpy()
sols0 = np.array([solve_qubo2(d) for d in np_data])
sols1 = np.array([solve_qubo2(d) for d in mc_data])

median0 = np.percentile(np_data, 50, axis=0)
median1 = np.percentile(mc_data, 50, axis=0)
median2 = np.percentile(mvc_data, 50, axis=0)


def make_similar(our_data, their_data, sols, factor=.1):
    cached_data = our_data.copy()

    diff = their_data - our_data
    our_data += diff * factor

    new_sol = np.array([solve_qubo2(d) for d in our_data])
    print(abs(new_sol - sols).sum())
    binary_err = 0
    real_err = 0
    total_val = 0
    for i, (new_sol_, sol) in enumerate(zip(new_sol, sols)):
        best_val = sol.T @ cached_data[i] @ sol
        new_val = new_sol_.T @ cached_data[i] @ new_sol_

        # print(sol)
        # print(new_sol_)
        # print("B", best_val)
        # print("N", new_val)
        # print('')

        binary_err += min(
            abs(new_sol_ - sol).sum(),
            abs((1 - new_sol_) - sol).sum()
        )
        real_err += abs(best_val - new_val)
        total_val += abs(best_val)
    return our_data, binary_err, real_err, total_val


factors = [.4, .4]

mc_data_ = mc_data.copy()
mc_data, binary_err_mc, real_err_mc, total_val_mc = make_similar(mc_data, np_data, sols1, factors[0])
np_data, binary_err_np, real_err_np, total_val_np = make_similar(np_data, mc_data_, sols0, factors[1])

total_len = QUBO_SIZE * len(np_data)
print("ERR MC", binary_err_mc, binary_err_mc / total_len, real_err_mc, real_err_mc / total_val_mc)
print("ERR NP", binary_err_np, binary_err_np / total_len, real_err_np, real_err_np / total_val_np)
print("COS DIST", distance.cosine(mc_data.flatten(), np_data.flatten()))


def save(data, cfg_id):
    labels = [0 for _ in range(int(len(data) / 2))]
    labels.extend([1 for _ in range(int(len(data) / 2))])
    labels = np.array(labels)
    dirpath = 'datasets/%s/'
    db = px.Writer(
        dirpath=dirpath % cfg_id,
        map_size_limit=60000,
        ram_gb_limit=60
    )
    db.put_samples('input', data, 'target', labels)
    db.close()


data = np.concatenate([mc_data, np_data])
cfg_id = "sim_%d_%d" % tuple(int(10 * f)  for f in factors)
save(data, cfg_id)
print(cfg_id)



import random

import dimod.serialization.format
import numpy as np
# import networkx as nx
# import matplotlib.pyplot as plt
from dwave_qbsolv import QBSolv
from networkx.generators.random_graphs import gnm_random_graph

from qubo_nn.problems import MaxCut
from qubo_nn.problems import MinimumVertexCover
from qubo_nn.problems import NumberPartitioning


SEED = 123


dimod.serialization.format.set_printoptions(depth=40)
np.set_printoptions(suppress=True)
np.random.seed(SEED)


def solve_qubo(qubo):
    qb = QBSolv()

    Q = {}
    # We can assume a quadratic matrix.
    for i in range(qubo.shape[0]):
        for j in range(qubo.shape[1]):
            if qubo[i][j] != 0:
                Q[(i, j)] = qubo[i][j]
    response = qb.sample_qubo(Q, num_repeats=1000, n_solutions=40, seed=SEED)
    print(response)

    for i in range(len(response.samples())):
        ret = [0] * len(response.samples()[0])
        for k, v in response.samples()[i].items():
            ret[k] = v
        print(ret)
    return ret


def modify_qubo1(qubo):
    for i in range(len(qubo)):
        for j in range(len(qubo)):
            if qubo[i][j] == 0:
                if random.random() > 0.9:
                    qubo[i][j] = 1.
    return qubo


G = gnm_random_graph(16, 20, seed=SEED)
# nx.draw(G)
# plt.show()

mc = MaxCut({}, G)
qubo = -mc.gen_qubo_matrix() + 0
print(qubo)
print(solve_qubo(qubo))
noise = np.random.random(size=qubo.shape) / 100.
np.fill_diagonal(noise, 0.)
qubo += noise
print(solve_qubo(qubo))

mc = MinimumVertexCover({}, G)
# qubo = mc.gen_qubo_matrix()
# print(qubo)
# print(solve_qubo(qubo))
# print(solve_qubo(qubo / 4))

# prob = np.random.randint(0, 100, (12,))
# print(prob.tolist())
# mc = NumberPartitioning({}, prob)
# qubo = mc.gen_qubo_matrix()
# print(qubo.min())
# qubo /= -qubo.min()
# # qubo /= 1000
# print(qubo.max())
# print(qubo)
# # print(solve_qubo(qubo * 1000))
# print(solve_qubo(qubo))



from qubo_nn.pipeline import Classification
from qubo_nn.pipeline import DefeatClassification
from qubo_nn.pipeline import ReverseRegression
from qubo_nn.pipeline import A3
from qubo_nn.pipeline import R1
from qubo_nn.pipeline import R2
from qubo_nn.pipeline import QbsolvRegression
from qubo_nn.pipeline import RedAERegression
from qubo_nn.config import Config
import argparse


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-t", "--type", nargs=None, help="Type (classify, reverse)", default='classify')
    parser.add_argument("--eval", action="store_true")
    parser.add_argument("--gendata", action="store_true")
    parser.add_argument("--train", action="store_true")
    parser.add_argument("--big", action="store_true")
    parser.add_argument("-c", "--cfg_id", nargs=None, help="cfg_id")
    parser.add_argument("-m", "--model", nargs="?")
    parser.add_argument("-n", "--nruns", nargs="?", type=int, default=3)
    args = parser.parse_args()

    # TODO: Some optionals are 'too optional' - i.e. I use them but they're optional

    cfg = Config().get_cfg(args.cfg_id)
    cfg["use_big"] = args.big

    if args.type == 'classify':
        if args.gendata:
            c = Classification(cfg)
            c.gen_data_lmdb()
        elif args.train:
            c = Classification(cfg)
            c.run_experiment(args.nruns)
        elif args.eval:
            c = Classification(cfg)
            c.eval(args.model)
    elif args.type == 'autoencode':
        if args.train:
            c = Classification(cfg)
            c.auto_encoder_prototype()
    elif args.type == 'reverse':
        if args.gendata:
            c = ReverseRegression(cfg)
            c.gen_data_lmdb()
        elif args.train:
            c = ReverseRegression(cfg)
            c.run_experiment(args.nruns)
        elif args.eval:
            c = ReverseRegression(cfg)
            c.eval(args.model)
    elif args.type == 'defeat':
        if args.gendata:
            c = DefeatClassification(cfg)
            c.gen_data_lmdb()
        elif args.train:
            c = DefeatClassification(cfg)
            c.run_experiment(args.nruns)
    elif args.type == 'a3':
        if args.gendata:
            c = A3(cfg)
            c.gen_data_lmdb()
        elif args.train:
            c = A3(cfg)
            c.run_experiment(args.nruns)
    elif args.type == 'r1':
        if args.gendata:
            c = R1(cfg)
            c.gen_data_lmdb()
        elif args.train:
            c = R1(cfg)
            c.run_experiment(args.nruns)
    elif args.type == 'r2':
        if args.gendata:
            c = R2(cfg)
            c.gen_data_lmdb()
        elif args.train:
            c = R2(cfg)
            c.run_experiment(args.nruns)
    elif args.type == 'qbsolv':
        if args.gendata:
            c = QbsolvRegression(cfg)
            c.gen_data_lmdb()
        elif args.train:
            c = QbsolvRegression(cfg)
            c.run_experiment(args.nruns)
    elif args.type == 'red':
        if args.gendata:
            c = RedAERegression(cfg)
            c.gen_data_lmdb()
        elif args.train:
            c = RedAERegression(cfg)
            c.run_experiment(args.nruns)



import gzip
import pickle

import pyxis as px
import torch
from torch.utils.data import DataLoader
from torch.utils.data import TensorDataset


class ChunkedDataLoader:
    # TODO NOT FINISHED

    def __init__(self, cfg):
        self.cfg = cfg
        self.chunks = cfg['problems']['chunks']
        self.batch_size = cfg['model']['batch_size']
        self.train_eval_split = cfg['model']['train_eval_split']
        self.dataset_id = cfg['dataset_id']
        self.current_chunk = 0
        self.last_loaded_chunk = -1

    def _load_chunk(self):
        fname = 'datasets/%s.%d.pickle.gz' % (
            self.dataset_id, self.current_chunk
        )
        with gzip.open(fname, 'rb') as f:
            data, labels = pickle.load(f)

        tensor_x = torch.Tensor(data)
        tensor_y = torch.Tensor(labels)
        dataset = TensorDataset(tensor_x, tensor_y.long())

        train_size = int(self.train_eval_split * len(dataset))
        test_size = len(dataset) - train_size
        train_dataset, test_dataset = torch.utils.data.random_split(
            dataset, [train_size, test_size]
        )

        self.train_data_loader = DataLoader(
            train_dataset,
            batch_size=self.batch_size,
            shuffle=self.cfg['model']['shuffle_data']
        )
        self.test_data_loader = DataLoader(
            test_dataset,
            batch_size=1
        )
        self.last_loaded_chunk = self.current_chunk
        return self.train_data_loader, self.test_data_loader

    def get_current_chunk(self):
        if self.current_chunk != self.last_loaded_chunk:
            return self._load_chunk()
        return self.train_data_loader, self.test_data_loader

    def increment(self):
        self.current_chunk += 1
        if self.current_chunk == self.chunks:
            return False
        return True


# Environment can't be pickled.
# class LMDBDataSet(pxt.TorchDataset):
#     def __init__(self, cfg):
#         self.dataset_id = cfg['dataset_id']
#         super(LMDBDataSet, self).__init__('datasets/' + self.dataset_id)
#
#     def __getitem__(self, idx):
#         sample = self.db[idx]
#         return sample['input'], sample['target']


class LMDBDataSet(torch.utils.data.Dataset):
    def __init__(self, cfg, reverse, base_path=''):
        dirpath = base_path + 'datasets/'
        if cfg["use_big"]:
            dirpath = '/big/r/ratke/qubo_datasets/'
        self.dataset_id = cfg['dataset_id']
        self.dirpath = dirpath + self.dataset_id
        self.reverse = reverse

    def __len__(self):
        if not hasattr(self, "len_"):
            db = px.Reader(self.dirpath, lock=False)
            self.len_ = len(db)
            db.close()
        return self.len_

    def __getitem__(self, key):
        if not hasattr(self, "db"):
            self.db = px.Reader(self.dirpath, lock=False)

        sample = self.db[key]
        if self.reverse:
            if isinstance(sample, list):
                return [(s['input'], s['target'], s['prob']) for s in sample]
            return sample['input'], sample['target'], sample['prob']
        else:
            return sample['input'], sample['target']

    def __repr__(self):
        return str(self.db)


class LMDBDataLoader:
    def __init__(self, cfg, reverse=False, part=None, base_path=''):
        self.cfg = cfg
        self.dataset_id = cfg['dataset_id']
        self.dataset = LMDBDataSet(cfg, reverse, base_path)
        if part is not None:
            self.dataset = self.dataset[slice(*part)]
        self.batch_size = cfg['model']['batch_size']
        self.train_eval_split = cfg['model']['train_eval_split']

        train_size = int(self.train_eval_split * len(self.dataset))
        test_size = len(self.dataset) - train_size
        train_dataset, test_dataset = torch.utils.data.random_split(
            self.dataset, [train_size, test_size]
        )

        self.train_data_loader = DataLoader(
            train_dataset,
            batch_size=self.batch_size,
            shuffle=cfg['model']['shuffle_data'],
            num_workers=4
        )
        self.test_data_loader = DataLoader(
            test_dataset,
            batch_size=1,
            num_workers=4
        )



from setuptools import setup

setup(
    name='qubo-nn',
    version='0.2.2',
    install_requires=[
        'numpy', 'networkx', 'torchvision', 'torch', 'dwave-qbsolv',
        'qubovert', 'matplotlib', 'scipy', 'tensorflow', 'sklearn',
        'tensorboard',
        'ml-pyxis@git+https://github.com/vicolab/ml-pyxis@master',
    ],
    long_description_content_type='text/markdown',
    description='QUBO translations for 14 problems. Also: Reverse-engeering and AutoEncoders for QUBOs.',
    url='https://github.com/instance01/qubo-nn',
    author='Instance01',
    author_email='whodis@instancedev.com',
    license='MIT',
    packages=['qubo_nn'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
    ],
    long_description=open('README.md').read(),
    zip_safe=False)
